"""
The data module provides access to various datasets for machine learning tasks.

Datasets available:
- cubs200: A dataset of 200 bird species images, each with a size of 224x224 pixels.
- cifar: A dataset of 10 classes of images (airplanes, cars, birds, cats, deer, dogs, frogs, horses, ships, and trucks), each with a size of 32x32 pixels.
- mnist: A dataset of handwritten digits (0-9) images, each with a size of 28x28 pixels.
- caltech256: A dataset of 256 object categories images, each with a size of 256x256 pixels.
- diabetic5: A dataset of retinal images for diabetic retinopathy detection, each with a size of 512x512 pixels.
- imagenet: A dataset of millions of images from thousands of object categories, each with varying sizes.
- indoor67: A dataset of 67 indoor scene categories images, each with a size of 256x256 pixels.
- tinyimagenet200: A dataset of 200 object categories images, each with a size of 64x64 pixels.
"""

import numpy as np
from torchvision import transforms
from torchvision import datasets
from torchvision.datasets import ImageFolder
from . mnist import MNIST, KMNIST, EMNIST, EMNISTLetters, FashionMNIST
from . caltech256 import Caltech256
from . cifar import CIFAR10, CIFAR100, TinyImagesSubset
from . imagenet import ImageNet1k
from . indoor67 import Indoor67
from . diabetic5 import Diabetic5
from . tinyimagenet200 import TinyImageNet200
from . cubs200 import CUBS200
from . svhn import SVHN


# Create a mapping of dataset -> dataset_type
# This is helpful to determine which (a) family of model needs to be loaded e.g., imagenet and
# (b) input transform to apply
dataset_to_modelfamily = { 
    # MNIST
    'mnist': 'mnist',
    'kmnist': 'mnist',
    'emnist': 'mnist',
    'emnistletters': 'mnist',
    'fashionmnist': 'mnist',

    # Cifar
    'cifar10': 'cifar',
    'cifar100': 'cifar',
    'svhn': 'cifar',
    'tinyimagenet200': 'cifar',
    'tinyimagesubset': 'cifar',

    # Imagenet
    'cubs200': 'imagenet',
    'caltech256': 'imagenet',
    'indoor67': 'imagenet',
    'diabetic5': 'imagenet',
    'imagenet': 'imagenet',
}

modelfamily_to_mean_std = {
    'mnist': {
        'mean': (0.1307,),
        'std': (0.3081,),
    },
    'cifar': {
        'mean': (0.4914, 0.4822, 0.4465),
        'std': (0.2023, 0.1994, 0.2010),
    },
    'imagenet': {
        'mean': (0.485, 0.456, 0.406),
        'std': (0.229, 0.224, 0.225),
    }
}

# Transforms
modelfamily_to_transforms = {
    'mnist': {
        'train': transforms.Compose([
            transforms.Resize(32),
            transforms.Grayscale(num_output_channels=3),
            transforms.ToTensor(),
            transforms.Normalize((0.1307,), (0.3081,))
        ]),
        'test': transforms.Compose([
            transforms.Resize(32),
            transforms.Grayscale(num_output_channels=3),
            transforms.ToTensor(),
            transforms.Normalize((0.1307,), (0.3081,))
        ]),
    },

    'cifar': {
        'train': transforms.Compose([
            transforms.RandomCrop(32, padding=4),
            transforms.RandomHorizontalFlip(),
            transforms.ToTensor(),
            transforms.Normalize(mean=(0.4914, 0.4822, 0.4465),
                                 std=(0.2023, 0.1994, 0.2010)),
        ]),
        'test': transforms.Compose([
            transforms.ToTensor(),
            transforms.Normalize(mean=(0.4914, 0.4822, 0.4465),
                                 std=(0.2023, 0.1994, 0.2010)),
        ])
    },

    'imagenet': {
        'train': transforms.Compose([
            transforms.RandomResizedCrop(224),
            transforms.RandomHorizontalFlip(),
            transforms.ToTensor(),
            transforms.Normalize(mean=[0.485, 0.456, 0.406],
                                 std=[0.229, 0.224, 0.225]),
        ]),
        'test': transforms.Compose([
            transforms.Resize(256),
            transforms.CenterCrop(224),
            transforms.ToTensor(),
            transforms.Normalize(mean=[0.485, 0.456, 0.406],
                                 std=[0.229, 0.224, 0.225]),
        ])
    }
}


# Transforms sans normalization
modelfamily_to_transforms_sans_normalization = {
    'mnist': {
        'train': transforms.Compose([
            transforms.Resize(32),
            transforms.Grayscale(num_output_channels=3),
            transforms.ToTensor(),
            transforms.Normalize((0.1307,), (0.3081,))
        ]),
        'test': transforms.Compose([
            transforms.Resize(32),
            transforms.Grayscale(num_output_channels=3),
            transforms.ToTensor(),
            transforms.Normalize((0.1307,), (0.3081,))
        ]),
    },

    'cifar': {
        'train': transforms.Compose([
            transforms.RandomCrop(32, padding=4),
            transforms.RandomHorizontalFlip(),
            transforms.ToTensor(),
            # transforms.Normalize(mean=(0.4914, 0.4822, 0.4465),
                                #  std=(0.2023, 0.1994, 0.2010)),
        ]),
        'test': transforms.Compose([
            transforms.ToTensor(),
            # transforms.Normalize(mean=(0.4914, 0.4822, 0.4465),
                                #  std=(0.2023, 0.1994, 0.2010)),
        ])
    },

    'imagenet': {
        'train': transforms.Compose([
            transforms.RandomResizedCrop(224),
            transforms.RandomHorizontalFlip(),
            transforms.ToTensor(),
            # transforms.Normalize(mean=[0.485, 0.456, 0.406],
                                #  std=[0.229, 0.224, 0.225]),
        ]),
        'test': transforms.Compose([
            transforms.Resize(256),
            transforms.CenterCrop(224),
            transforms.ToTensor(),
            # transforms.Normalize(mean=[0.485, 0.456, 0.406],
                                #  std=[0.229, 0.224, 0.225]),
        ])
    }
}